 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#define	MAIN_PROGRAM

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <values.h>
#include <sys/mman.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <gl/image.h>

#include "data.h"
#include "rgba.h"

static int
	texture_colortable = 1,
	ElevatorWIN, W, H,
	mousex, mousey,
	lat_stride, lon_stride, lod=3,
	mousedown = 0,
	textured = 1,
	detailed = 1,
	wireframed = 0;

static float
	*RGBA = NULL,
	ALTMIN, ALTMAX,
	obsLat, obsLon, obsAlt,
	cos_head, sin_head,
	cos_pitch, sin_pitch,
	speed, delta_speed,
	roll, pitch, head;

#define	TEXTURE_WIDTH	1024
#define	TEXTURE_HEIGHT	1024
#define	BASE_WIDTH		512
#define	BASE_HEIGHT		512
#define	DETAIL_WIDTH	256
#define	DETAIL_HEIGHT	256

unsigned char Texture[TEXTURE_WIDTH*TEXTURE_HEIGHT];
unsigned char Base[BASE_WIDTH*BASE_HEIGHT];
unsigned char Detail[DETAIL_WIDTH*DETAIL_HEIGHT];

static void Redraw( void );
static void Idle( void );
static void Mouse( int, int, int, int );
static void Motion( int, int );
static void Reshape( int, int );
static void Key( unsigned char, int, int );

static void InitTexture( void );
static int ReadTerrain( char * );
void stripArray( int, int, int );
void AfficheTexturedTerrain( int, int, int );
void AfficheColoredTerrain( int, int, int );
int XCullDraw(
    float altmin, float altmax,
    int numlat, int lat_stride, float d_lat,
    int numlon, int lon_stride, float d_lon,
    void (*drawFunc)( int, int, int )
);

static void newLUTCallBack( int * );
void menuCallBack( int );

/*---------------------------------------------------------------------------*/
/*				MAIN
/*---------------------------------------------------------------------------*/
void
main( int argc, char *argv[] )
{
	glutInit( &argc, argv );

/*
	Local Initializations
	---------------------
*/
	ReadTerrain( NULL );

	obsLat	= ((float) (NUM_LATITUDE/2))*D_LATITUDE;
	obsLon	= ((float) (NUM_LONGITUDE/2))*D_LONGITUDE;
	obsAlt	= ALTMAX;
	roll = pitch = head = 0;
	speed = 10.;
	delta_speed = 0.;
	near_cull_dist = 5.;
	far_cull_dist = 7500.;
	
/*
	Graphics Initializations
	------------------------
*/
	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
	glutInitWindowPosition( 0, 0 );
	glutInitWindowSize( 850, 1000 );
	ElevatorWIN = glutCreateWindow( "terrain following" );

	glutDisplayFunc( Redraw );
	glutReshapeFunc( Reshape );
	glutMouseFunc( Mouse );
	glutIdleFunc(Idle);
	glutPassiveMotionFunc( Motion );
	glutMotionFunc( Motion );
	glutKeyboardFunc(Key);

	makePopupMenu();

	YACME_init(
		900, 471, 400, 545,
		256, &RGBA,
		newLUTCallBack
	);

	glutMainLoop();
}

/*---------------------------------------------------------------------------*/
/*	InitGraphique
/*---------------------------------------------------------------------------*/
static void
InitGraphique( void )
{
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_CULL_FACE );
	glCullFace( GL_BACK );

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60., 1., near_cull_dist, far_cull_dist );
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	InitTexture();

	glClearColor( .5, .5, 1., 0. );
	glClearDepth( 1. );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glutSwapBuffers();
}

/*---------------------------------------------------------------------------*/
/*	Redraw
/*---------------------------------------------------------------------------*/
static void
Redraw( void )
{
	float aux, alt;
	int i, j;

	glClearColor( .5, .5, 1., 0. );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	if (mousedown)
	{
		head  += (float)(2*mousex - W) / W;
		pitch += (float)(H - 2*mousey) / H;

#ifdef DEV
		fprintf( stderr, "\n\nobsLat = %f\n"
			"obsLon = %f\nobsAlt = %f\nhead = %f\npitch = %f\n\n",
			obsLat, obsLon, obsAlt, head, pitch );
#endif

		while (pitch < 0.) pitch += 360.;
		while (pitch > 360.) pitch -= 360.;
		while (head < 0.) head += 360.;
		while (head > 360.) head -= 360.;

		aux = DEG2RAD((float) head);
		cos_head = fcos( aux );
		sin_head = fsin( aux );
		aux = DEG2RAD((float) pitch);
		cos_pitch = fcos( aux );
		sin_pitch = fsin( aux );

		speed += delta_speed;
		obsLat -= speed*cos_head*cos_pitch;
		obsLon += speed*sin_head*cos_pitch;
		obsAlt -= speed*sin_pitch;
	}

/*
	Check against terrain's boundaries
*/
	i =(int) obsLat/D_LATITUDE;
	j =(int) obsLon/D_LONGITUDE;
	if ((i>=0) && (j>=0) && (i<NUM_LATITUDE) && (j<NUM_LONGITUDE))
	{
		alt = Terrain[i][j].alt;
		alt += RADADA;
		obsAlt = SUP( alt, obsAlt );
	}
	else if (mousedown)
	{
		obsLat += speed*cos_head*cos_pitch;
		obsLon -= speed*sin_head*cos_pitch;
		obsAlt += speed*sin_pitch;
	}

#if 0
/*
	Adaptative LOD - now user action
*/
	lod = SUP( obsAlt/(ALTMAX*.5), 1. );
#endif
	lat_stride = lod*LAT_STRIDE;
	lon_stride = lod*LON_STRIDE;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60., 1., lod*near_cull_dist, lod*far_cull_dist );
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glLoadIdentity();
	glRotatef( pitch, 1.0, 0.0, 0.0 );
	glRotatef( head, 0.0, 1.0, 0.0 );
	glTranslatef(-obsLon,-obsAlt,-obsLat);

	if (wireframed)
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	else
		glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

#ifdef VARRAY
	glVertexPointerEXT(
		3,							/* size=number of coordinates */
		GL_FLOAT,					/* type */
		sizeof(Elevation),			/* stride */
		NUM_LONGITUDE*NUM_LATITUDE,	/* static */
		&Terrain[0][0].lon
	);
	glColorPointerEXT(
		4,							/* size=number of coordinates */
		GL_UNSIGNED_BYTE,			/* type */
		sizeof(Elevation),			/* stride */
		NUM_LONGITUDE*NUM_LATITUDE,	/* static */
		&Terrain[0][0].c
	);
#endif
	if (textured)
	{
		if (texture_colortable)
			glEnable( GL_TEXTURE_COLOR_TABLE_SGI );

		if (!wireframed)
			glEnable( GL_TEXTURE_2D );

		if (detailed)
		{
    		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
				GL_LINEAR_MIPMAP_LINEAR);
    		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
				GL_LINEAR_DETAIL_SGIS);
		}
		else
		{
    		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}

#ifdef VARRAY
		glEnable( GL_VERTEX_ARRAY_EXT );
		XCullDraw( ALTMIN, ALTMAX,
			NUM_LATITUDE, lat_stride, D_LATITUDE,
			NUM_LONGITUDE, lon_stride, D_LONGITUDE,
			stripArray );
		glDisable( GL_VERTEX_ARRAY_EXT );
#else	/* VARRAY */
		XCullDraw( ALTMIN, ALTMAX,
			NUM_LATITUDE, lat_stride, D_LATITUDE,
			NUM_LONGITUDE, lon_stride, D_LONGITUDE,
			AfficheTexturedTerrain );
#endif	/* VARRAY */
		glDisable( GL_TEXTURE_COLOR_TABLE_SGI );
		glDisable( GL_TEXTURE_2D );
	}
	else
	{
		glDisable( GL_TEXTURE_2D );

#ifdef VARRAY
		glEnable( GL_VERTEX_ARRAY_EXT );
		glEnable( GL_COLOR_ARRAY_EXT );
		XCullDraw( ALTMIN, ALTMAX,
			NUM_LATITUDE, lat_stride, D_LATITUDE,
			NUM_LONGITUDE, lon_stride, D_LONGITUDE,
			stripArray );
		glDisable( GL_VERTEX_ARRAY_EXT );
		glDisable( GL_COLOR_ARRAY_EXT );
#else	/* VARRAY */
		XCullDraw( ALTMIN, ALTMAX,
			NUM_LATITUDE, lat_stride, D_LATITUDE,
			NUM_LONGITUDE, lon_stride, D_LONGITUDE,
			AfficheColoredTerrain );
#endif	/* VARRAY */
	}

	glutSwapBuffers();
}

/*---------------------------------------------------------------------------*/
/*	Idle
/*---------------------------------------------------------------------------*/
static void
Idle( void )
{
	glutSetWindow( ElevatorWIN );
	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Mouse
/*---------------------------------------------------------------------------*/
static void
Mouse( int button, int state, int x, int y )
{
	switch (button)
	{
		case GLUT_LEFT_BUTTON:
			if (mousedown)
				delta_speed = (state==GLUT_DOWN) ? -1. : 0.;
			else if (state == GLUT_DOWN)
				if (speed < 0.) speed = -speed;
			break;

		case GLUT_MIDDLE_BUTTON:
			if (mousedown)
				delta_speed = (state==GLUT_DOWN) ? 1. : 0.;
			else if (state == GLUT_DOWN)
				if (speed > 0.) speed = -speed;
			break;

		case GLUT_RIGHT_BUTTON:
			break;
	}
	mousedown += (state==GLUT_DOWN) ? 1 : -1;
	if (!mousedown) speed = delta_speed = 0;
}

/*---------------------------------------------------------------------------*/
/*	Motion
/*---------------------------------------------------------------------------*/
static void
Motion( int x, int y )
{
	mousex = x;
	mousey = y;
}

/*---------------------------------------------------------------------------*/
/*		Reshape
/*---------------------------------------------------------------------------*/
static void
Reshape( int w, int h )
{
	static int first = 1;
	glViewport( 0, 0, w, h );
	W = w;
	H = h;
	mousex = W/2;
	mousey = H/2;

	if (first)
	{
		first = 0;
		InitGraphique();
	}
}

/*---------------------------------------------------------------------------*/
/*		Key
/*---------------------------------------------------------------------------*/
static void
Key(unsigned char key, int x, int y)
{
	switch (key)
	{
		case 27: /* Escape */
			exit(0);

		case 'w':
		case 'W':
			wireframed = !wireframed;
			break;

		case 't':
		case 'T':
			textured = !textured;
			break;

		case 'd':
		case 'D':
			detailed = !detailed;
			InitTexture();
			break;

		case 'l':
			if (lod<16) lod++;
			break;

		case 'L':
			if (lod>1) lod--;
			break;
	}
}

/*---------------------------------------------------------------------------*/
/*			ReadTerrain
/*---------------------------------------------------------------------------*/
extern int ColorTerrain();
static int
ReadTerrain( char *filename )
{
	int		lat, lon;
	int		fd;
	float	altitude;
	float	*data;

	struct stat status;

	fprintf(stderr, "Reading Elevation data: ...", filename);

/*
	mmap'ing the data is SO MUCH FASTER !
*/
	fd = open( "../data/elevator.bin", O_RDONLY );
	fstat( fd, &status );
	data =
	(float *) mmap(NULL,(size_t)status.st_size,PROT_READ,MAP_SHARED,fd,0);

	ALTMAX = ALTMIN = data[0];
	for (lat=0; lat<NUM_LATITUDE; lat++)
	{
		for (lon=0; lon<NUM_LONGITUDE; lon++)
		{
			altitude = data[lat*NUM_LONGITUDE + lon];
			Terrain[lat][lon].lon = ((float) lon)*D_LONGITUDE;
			Terrain[lat][lon].alt = altitude;
			Terrain[lat][lon].lat = ((float) lat)*D_LATITUDE;
			ALTMAX = SUP( ALTMAX, altitude );
			ALTMIN = INF( ALTMIN, altitude );
		}
		if ( !(lat%10) ) fprintf(stderr, ".");
	}
	fprintf(stderr, "Done\n");
	close( fd );

	ColorTerrain();
}

/*---------------------------------------------------------------------------*/
/*			AfficheTerrain
/*---------------------------------------------------------------------------*/
/*
 	Elevation data are drawn using longitude-aligned tmeshes, as follow:

		lat	^
			|
			|
   			+	1 3 5 ...(2n+1)
D_LATITUDE	|	|\|\|\	    |
   		  	+	0 2 4 ... (2n)
			|	
			+-------+-+--------------------------------------------------> lon
				D_LONGITUDE

*/

#ifdef VARRAY
void
stripArray( int lat, int first, int last )
{
	int lon;
	int dlat = lat_stride*NUM_LONGITUDE;

	first += lat*NUM_LONGITUDE;
	last += lat*NUM_LONGITUDE;
	
	glColor3f( 1., 1., 1. );
	glBegin( GL_TRIANGLE_STRIP );
	for (lon=first; lon<=last; lon+=lon_stride)
	{
		glArrayElementEXT( lon );
		glArrayElementEXT( lon + dlat );
	}
	glEnd();

} /* stripArray */

#endif	/* VARRAY */

void
AfficheTexturedTerrain( int lat, int first, int last )
{
	int lon;

	glColor3f( 1., 1., 1. );
	glBegin( GL_TRIANGLE_STRIP );
	for (lon=first; lon<=last; lon+=lon_stride)
	{
		glVertex3fv((GLfloat *) &Terrain[lat][lon].lon);
 		glVertex3fv((GLfloat *) &Terrain[lat+lat_stride][lon].lon);
	}
	glEnd();

} /* AfficheTexturedTerrain */

void
AfficheColoredTerrain( int lat, int first, int last )
{
	int lon;

	glBegin( GL_TRIANGLE_STRIP );
	for (lon=first; lon<=last; lon+=lon_stride)
	{
		glColor4ubv((GLubyte *) &Terrain[lat][lon].c );
		glVertex3fv((GLfloat *) &Terrain[lat][lon].lon);
		glColor4ubv((GLubyte *) &Terrain[lat+lat_stride][lon].c);
 		glVertex3fv((GLfloat *) &Terrain[lat+lat_stride][lon].lon);
	}
	glEnd();

} /* AfficheColoredTerrain */

/*---------------------------------------------------------------------------*/
/*	InitTexture
/*---------------------------------------------------------------------------*/
static void
InitTexture( void )
{
	static GLfloat sparam[] =
		{ 0., 0., 1./((float)NUM_LATITUDE*D_LATITUDE), 0. };
	static GLfloat tparam[] =
		{ -1./((float)NUM_LONGITUDE*D_LONGITUDE), 0., 0., 1. };
	static int first = 1;

	if (first)
	{
		int i, j;
		short rowbuf[1024];
		unsigned char *texptr;
		IMAGE *image;

#define	READIMAGE( file, texture )\
		\
		image = iopen(file,"rw");\
		texptr = texture;\
		for (i=0; i<image->ysize; i++)\
		{\
			getrow( image, rowbuf, i, 0 );\
			for (j=0; j<image->xsize; j++)\
			{\
				*texptr++ = (char) rowbuf[j];\
			}\
		}\
		iclose( image );

		READIMAGE( "../data/detailed.bw", Base );
		READIMAGE( "../data/terrain.bw", Detail );
		READIMAGE( "../data/elevator.bw", Texture );

		first = 0;
	}

	if (detailed)
	{
		gluBuild2DMipmaps(
			GL_TEXTURE_2D,		/* target */
			1,					/* component */
			BASE_WIDTH,
			BASE_HEIGHT,
			GL_LUMINANCE,		/* EXTERNAL format */
			GL_UNSIGNED_BYTE,	/* EXTERNAL type */
			Base				/* data */
		);
		glTexImage2D(
			GL_DETAIL_TEXTURE_2D_SGIS,
			0,
			1,
			DETAIL_WIDTH,
			DETAIL_HEIGHT,
			0,
			GL_LUMINANCE,
			GL_UNSIGNED_BYTE,
			Detail
		);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_LINEAR_DETAIL_SGIS );
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_LINEAR);
	    glTexParameteri(GL_TEXTURE_2D, GL_DETAIL_TEXTURE_LEVEL_SGIS, -1 );
	    glTexParameteri(GL_TEXTURE_2D, GL_DETAIL_TEXTURE_MODE_SGIS, GL_ADD );

	}
	else	/* NOT detailed */
	{
		glTexImage2D(
			GL_TEXTURE_2D,
			0,
			GL_RGB8_EXT,
			TEXTURE_WIDTH,
			TEXTURE_HEIGHT,
			0,
			GL_LUMINANCE,
			GL_UNSIGNED_BYTE,
			Texture
		);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	}
   	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	glEnable( GL_TEXTURE_2D );

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glMatrixMode( GL_MODELVIEW );

	glTexGeni( GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR );
	glTexGenfv( GL_S, GL_OBJECT_PLANE, sparam );
	glTexGeni( GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR );
	glTexGenfv( GL_T, GL_OBJECT_PLANE, tparam );
	glEnable( GL_TEXTURE_GEN_S );
	glEnable( GL_TEXTURE_GEN_T );
}

/*---------------------------------------------------------------------------*/
/*	newLUTCallBack
/*---------------------------------------------------------------------------*/
static void
newLUTCallBack( int *hasChanged )
{
	int oldwin = glutGetWindow();

	glutSetWindow( ElevatorWIN );
	glColorTableSGI(
		GL_TEXTURE_COLOR_TABLE_SGI,
		GL_RGB,
		256,
		GL_RGBA,
		GL_FLOAT,
		RGBA
	); 
/*
	Make a Detail Func out of ALPHA values
	--------------------------------------
*/
	if (hasChanged[3])
	{
		int i;
		GLfloat detailCurve[16];

		for (i=0; i<8; i++)
		{
			detailCurve[i*2] = -(float)i;
			detailCurve[i*2+1] = RGBA[128*i + 3];
		}
		glDetailTexFuncSGIS( GL_TEXTURE_2D, 8, detailCurve );
	}
	glutSetWindow( oldwin );
}

/*---------------------------------------------------------------------------*/
/*	makePopupMenu
/*---------------------------------------------------------------------------*/
#define	DRAW_WIRE_ITEM		1
#define	DRAW_SHADED_ITEM	2
#define	DRAW_TEX_ITEM		3
#define	DRAW_DETAIL_ITEM	4

#define	LUT_ON_ITEM		11
#define	LUT_OFF_ITEM	12

#define LOD_UP_ITEM		21
#define	LOD_DOWN_ITEM	22

makePopupMenu()
{
	int draw_menu, lut_menu, lod_menu, MAIN_menu;

	draw_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "Detailed->Wireframe", DRAW_WIRE_ITEM );

	lut_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "On->Off", LUT_OFF_ITEM );

	lod_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "Up", LOD_UP_ITEM );
	glutAddMenuEntry( "Down", LOD_DOWN_ITEM );

	MAIN_menu = glutCreateMenu( menuCallBack );
	glutAddSubMenu( "Draw Style", draw_menu );
	glutAddSubMenu( "Texture LUT", lut_menu );
	glutAddSubMenu( "Tesselation", lod_menu );

	glutAttachMenu( GLUT_RIGHT_BUTTON );
}

void
menuCallBack( int item )
{
	switch (item)
	{
		case DRAW_WIRE_ITEM:
			glutChangeToMenuEntry( 1, "Wireframe->Shaded", DRAW_SHADED_ITEM );
			wireframed = GL_TRUE;
			textured = GL_FALSE;
			detailed = GL_FALSE;
			break;

		case DRAW_SHADED_ITEM:
			glutChangeToMenuEntry( 1, "Shaded->Textured", DRAW_TEX_ITEM );
			wireframed = GL_FALSE;
			textured = GL_FALSE;
			detailed = GL_FALSE;
			break;

		case DRAW_TEX_ITEM:
			glutChangeToMenuEntry( 1, "Textured->Detailed", DRAW_DETAIL_ITEM );
			wireframed = GL_FALSE;
			textured = GL_TRUE;
			detailed = GL_FALSE;
			InitTexture();
			break;

		case DRAW_DETAIL_ITEM:
			glutChangeToMenuEntry( 1, "Detailed->Wireframe", DRAW_WIRE_ITEM );
			wireframed = GL_FALSE;
			textured = GL_TRUE;
			detailed = GL_TRUE;
			InitTexture();
			break;

		case LUT_ON_ITEM:
			glutChangeToMenuEntry( 1, "On->Off", LUT_OFF_ITEM );
			texture_colortable = GL_TRUE;
			break;

		case LUT_OFF_ITEM:
			glutChangeToMenuEntry( 1, "Off->On", LUT_ON_ITEM );
			texture_colortable = GL_FALSE;
			break;

		case LOD_UP_ITEM:
			if (lod<16) lod++;
			break;

		case LOD_DOWN_ITEM:
			if (lod>1) lod--;
			break;
	}
	glutPostRedisplay();
}
