 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <values.h>
#include <sys/mman.h>

#include "data.h"

/*
 Lighting vector
*/
#define LX      0.0f
#define LY      .447213595
#define LZ      .894427190

#define CHANCE      ( ((float) random()) / ((float) MAXLONG) )
#define RND(x)      ( ((float) x) * CHANCE)

#define GAMMA				1.00001
#define SHADOW_INTENSITY	.5

int GetAltColor(float *rr, float *gg, float *bb, float alt);

/*---------------------------------------------------------------------------*/
/*			ColorTerrain
/*---------------------------------------------------------------------------*/
int
ColorTerrain( void )
{
	int	i, j;
	float	alt, rr, gg, bb, N[3], D=1.;
	union {
		unsigned char cpste[4];
		unsigned long packed;
	} couleur;

	for (i=0; i<NUM_LATITUDE; i++)
	{
		for (j=0; j<NUM_LONGITUDE; j++)
		{
			alt = Terrain[i][j].alt;

#ifdef ALTNOISE
			alt *= (0.875 + RND(.25));
#endif
			GetAltColor(&rr, &gg, &bb, alt);
/*
			compute pseudo-normal to terrain
*/
#ifdef ALTSHADOW
			if ((i>0) && (i+1)<NUM_LATITUDE)
				N[0] =  Terrain[i-1][j].alt - Terrain[i+1][j].alt;
			else
				N[0] = Terrain[i][j].alt;
			if ((j>0) && (j+1)<NUM_LONGITUDE)
				N[1] =  Terrain[i][j-1].alt - Terrain[i][j+1].alt;
			else
				N[1] = Terrain[i][j].alt;
			N[2] = 100.;

			D = 1.0 / sqrtf(N[0]*N[0] + N[1]*N[1] + N[2]*N[2]);

			N[0] *= D;
			N[1] *= D;
			N[2] *= D;
/*
			perform diffuse lighting of terrain
*/
			D = N[0]*LX + N[1]*LY + N[2]*LZ;
			D *= 1.2;
/* #ifdef ALTNOISE */
			if(!((int) RND(3))) D *= 1.25f;
/* #endif */
			D = SUP(D,SHADOW_INTENSITY);
/*
			brighten terrain on lighted side
*/
#endif	/* ALTSHADOW */

			D *= GAMMA;
			rr *= D*255.;
			gg *= D*255.;
			bb *= D*255.;

			rr = INF( SUP(rr,0.), 255. );
			gg = INF( SUP(gg,0.), 255. );
			bb = INF( SUP(bb,0.), 255. );
			couleur.cpste[0] = (char) rr;
			couleur.cpste[1] = (char) gg;
			couleur.cpste[2] = (char) bb;
			couleur.cpste[3] = (char) 255;

			Terrain[i][j].c = couleur.packed;
		}
	}
}

/*---------------------------------------------------------------------------*/
/*			GetAltColor
/*---------------------------------------------------------------------------*/
/*
 Computes RGB components relatively to a given altitude
*/
int
GetAltColor(float *rr, float *gg, float *bb, float alt)
{
	float coeff;
/*
	green to redish lowlands
*/
	if (alt <1000.)
	{

		coeff = alt/1000.;
		*rr = .25*coeff;
		*gg = .4*(1.0 - coeff) + .14;
		*bb = .1;
/*
	rocky hillsides
*/
	}
	else if (alt < 1500.)
	{
		coeff = alt/1500.;
		*rr = .25*(1.-coeff) + coeff*.6;
		*gg = .2*(1.-coeff) + coeff*.5;
		*bb = .05*(1.-coeff) + coeff*.2;
/*
	 increasingly white snow
*/
	}
	else
	{
		*rr = alt/2000.;
		*gg = *rr;
		*bb = *rr;
	}
}
